"""DB Models for cloudstorage plugin."""

from datetime import datetime

import ckan.model as model
import ckan.model.meta as meta
from ckan.model.domain_object import DomainObject
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    Numeric,
    UnicodeText,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref, relationship

Base = declarative_base()
metadata = Base.metadata


def drop_tables():
    """Drop all tables."""
    metadata.drop_all(model.meta.engine)


def create_tables():
    """Create all tables."""
    metadata.create_all(model.meta.engine)


class MultipartPart(Base, DomainObject):
    """MultiPart table."""

    __tablename__ = "cloudstorage_multipart_part"

    def __init__(self, n, etag, upload):
        """Table init."""
        self.n = n
        self.etag = etag
        self.upload = upload

    n = Column(Integer, primary_key=True)
    etag = Column(UnicodeText, primary_key=True)
    upload_id = Column(
        UnicodeText, ForeignKey("cloudstorage_multipart_upload.id"), primary_key=True
    )
    upload = relationship(
        "MultipartUpload",
        backref=backref("parts", cascade="delete, delete-orphan"),
        single_parent=True,
    )


class MultipartUpload(Base, DomainObject):
    """MultipartUpload table."""

    __tablename__ = "cloudstorage_multipart_upload"

    def __init__(self, id, resource_id, name, size, original_name, user_id):
        """Table init."""
        self.id = id
        self.resource_id = resource_id
        self.name = name
        self.size = size
        self.original_name = original_name
        self.user_id = user_id

    @classmethod
    def resource_uploads(cls, resource_id):
        """Get all resource uploads for a resource."""
        query = meta.Session.query(cls).filter_by(resource_id=resource_id)
        return query

    id = Column(UnicodeText, primary_key=True)
    resource_id = Column(UnicodeText)
    name = Column(UnicodeText)
    initiated = Column(DateTime, default=datetime.utcnow)
    size = Column(Numeric)
    original_name = Column(UnicodeText)
    user_id = Column(UnicodeText)
    upload_complete = Column(Boolean, default=False)
